"""
Provide functions to interact with persistent data storage.
"""
import re
import typing

import asyncpg
import databases
import databases.core
import fastapi
import pydantic
import sqlalchemy
from asyncpg.exceptions import UniqueViolationError
from fastapi.exceptions import HTTPException
from loguru import logger
from sqlalchemy import Column, or_
from sqlalchemy.sql.expression import BooleanClauseList, UnaryExpression
from starlette import status
from yarl import URL

from jobbergate_api.config import settings

INTEGRITY_CHECK_EXCEPTIONS = (UniqueViolationError,)


def build_db_url(force_test: bool = False) -> str:
    """
    Build a database url based on settings.

    If the ``DEPLOY_ENV`` setting is "TEST" or if the ``force_test`` is passed, build from the test database
    settings.
    """
    is_test = force_test or settings.DEPLOY_ENV.lower() == "test"
    prefix = "TEST_" if is_test else ""

    return str(
        URL.build(
            scheme="postgresql",
            user=getattr(settings, f"{prefix}DATABASE_USER"),
            password=getattr(settings, f"{prefix}DATABASE_PSWD"),
            host=getattr(settings, f"{prefix}DATABASE_HOST"),
            port=getattr(settings, f"{prefix}DATABASE_PORT"),
            path="/{}".format(getattr(settings, f"{prefix}DATABASE_NAME")),
        )
    )


database = databases.Database(build_db_url(), force_rollback=settings.DEPLOY_ENV.lower() == "test")


def render_sql(query) -> str:
    """
    Render a sqlalchemy query into a string for debugging.
    """
    return query.compile(dialect=database._backend._dialect, compile_kwargs={"literal_binds": True})


def search_clause(
    search_terms: str,
    searchable_fields: typing.List[Column],
) -> BooleanClauseList:
    """
    Create search clause across searchable fields with search terms.
    """
    return or_(*[field.ilike(f"%{term}%") for field in searchable_fields for term in search_terms.split()])


def sort_clause(
    sort_field: str,
    sortable_fields: typing.List[Column],
    sort_ascending: bool,
) -> typing.Union[Column, UnaryExpression]:
    """
    Create a sort clause given a sort field, the list of sortable fields, and a sort_ascending flag.
    """
    sort_field_names = [f.name for f in sortable_fields]
    try:
        index = sort_field_names.index(sort_field)
    except ValueError:
        raise HTTPException(
            status_code=status.HTTP_400_BAD_REQUEST,
            detail=f"Invalid sorting column requested: {sort_field}. Must be one of {sort_field_names}",
        )
    sort_column: typing.Union[Column, UnaryExpression] = sortable_fields[index]
    if not sort_ascending:
        sort_column = sort_column.desc()
    return sort_column


def handle_fk_error(
    _: fastapi.Request,
    err: asyncpg.exceptions.ForeignKeyViolationError,
):
    """
    Unpack metadata from a ForeignKeyViolationError and return a 409 response.
    """
    FK_DETAIL_RX = r"DETAIL:  Key \(id\)=\((?P<pk_id>\d+)\) is still referenced from table \"(?P<table>\w+)\""
    matches = re.search(FK_DETAIL_RX, str(err), re.MULTILINE)
    (table, pk_id) = (None, None)
    if matches:
        table = matches.group("table")
        pk_id = matches.group("pk_id")

    logger.error(f"Delete failed due to foreign-key constraint: {table=} {pk_id=}")

    return fastapi.responses.JSONResponse(
        status_code=fastapi.status.HTTP_409_CONFLICT,
        content=dict(
            detail=dict(
                message="Delete failed due to foreign-key constraint",
                table=table,
                pk_id=pk_id,
            ),
        ),
    )


T = typing.TypeVar("T", bound=pydantic.BaseModel)


async def fetch_instance(id: int, table: sqlalchemy.Table, model: typing.Type[T]) -> T:
    """
    Fetch a single frow from a table by its id and unpack it into a response model.
    """
    query = table.select(table.c.id == id)
    result = await database.fetch_one(query)
    if result is None:
        raise HTTPException(status.HTTP_404_NOT_FOUND, f"Could not find {table.name} instance with id {id}")
    return model.parse_obj(result)



"""
Provide a module that describes permissions in the API.
"""

from enum import Enum


class Permissions(str, Enum):
    """
    Describe the permissions that may be used for protecting Jobbergate routes.
    """

    APPLICATIONS_VIEW = "jobbergate:applications:view"
    APPLICATIONS_EDIT = "jobbergate:applications:edit"
    JOB_SCRIPTS_VIEW = "jobbergate:job-scripts:view"
    JOB_SCRIPTS_EDIT = "jobbergate:job-scripts:edit"
    JOB_SUBMISSIONS_VIEW = "jobbergate:job-submissions:view"
    JOB_SUBMISSIONS_EDIT = "jobbergate:job-submissions:edit"



"""
Convenience fixtures for the job scripts tests.
"""

from pytest import fixture


@fixture
def param_dict():
    """
    Provide a fixture that returns a param dict.
    """
    return {
        "application_config": {"job_name": "rats", "partitions": ["debug", "partition1"]},
        "jobbergate_config": {
            "default_template": "test_job_script.sh",
            "job_name": "rats",
            "output_directory": ".",
            "partition": "debug",
            "supporting_files": ["test_job_script.sh"],
            "supporting_files_output_name": {"test_job_script.sh": ["support_file_b.py"]},
            "template_files": ["templates/test_job_script.sh"],
        },
    }


@fixture
def param_dict_flat():
    """
    Provide a fixture tha returns a flattened pram dict.
    """
    return {
        "job_name": "rats",
        "partitions": ["debug", "partition1"],
        "default_template": "test_job_script.sh",
        "output_directory": ".",
        "partition": "debug",
        "supporting_files": ["test_job_script.sh"],
        "supporting_files_output_name": {"test_job_script.sh": ["support_file_b.py"]},
        "template_files": ["templates/test_job_script.sh"],
    }


@fixture
def template_files(dummy_template):
    """
    Provide a fixture that creates test template files.
    """
    return {"application.sh": dummy_template}



"""
Validation methods for the uploaded files.
"""

from ast import parse as ast_parse
from collections import Counter
from functools import wraps
from pathlib import PurePath
from textwrap import dedent
from traceback import format_tb
from typing import Callable, Dict, List, Union

from buzz import Buzz, DoExceptParams, require_condition
from fastapi import HTTPException, UploadFile, status
from jinja2 import Template
from loguru import logger
from yaml import safe_load as yaml_safe_load

from jobbergate_api.apps.applications.schemas import ApplicationConfig


class UploadedFilesValidationError(Buzz):
    """Raise exception when facing any validation error on the uploaded files."""


def log_error_and_raise_http_error(params: DoExceptParams):
    """
    Work with ``UploadedFilesValidationError.handle_errors`` when something goes wrong.

    In case any error occurs in the context manager, the error is logged and
    a HTTPException is raised.

    :param DoExceptParams params: Params from handle_errors.
    :raises HTTPException: Code 422, unprocessable entity.
    """
    message_template = dedent(
        """
        {final_message}
        Error:
        ______
        {err}
        Traceback:
        ----------
        {trace}
        """
    )

    message = message_template.format(
        final_message=params.final_message,
        err=str(params.err),
        trace="\n".join(format_tb(params.trace)),
    )

    logger.error(message)

    raise HTTPException(status_code=status.HTTP_422_UNPROCESSABLE_ENTITY, detail=params.final_message)


def get_suffix(file: UploadFile) -> str:
    """
    Get the suffix (file extension) of an UploadFile.

    :param UploadFile file: Target file.
    :return str: File extension.
    """
    return PurePath(file.filename).suffix


CheckUploadedFilesEquation = Callable[[List[UploadFile]], None]
"""
Type alias describing the function signature used to check the uploaded files.

The functions are expected to raise UploadedFilesValidationError if any error is detected.
"""

check_uploaded_files_dispatch: List[CheckUploadedFilesEquation] = []
"""List the function used to check the uploaded files."""


def perform_all_checks_on_uploaded_files(file_list: List[UploadFile]):
    """
    Loop though the check-list for uploaded files specifications.

    Individual checks will raise UploadedFilesValidationError if the business
    rules are not met.

    :param List[UploadFile] file_list: Upload file list.
    """
    logger.debug("Preparing to perform all checks on the uploaded files")
    with UploadedFilesValidationError.handle_errors(
        "The uploaded files are invalid",
        do_except=log_error_and_raise_http_error,
        re_raise=False,
    ):
        for check in check_uploaded_files_dispatch:
            check(file_list)


def register_check_uploaded_files(checker: CheckUploadedFilesEquation) -> CheckUploadedFilesEquation:
    """
    Use this decorator to register functions to check the uploaded files.

    It creates a new entry on ``check_uploaded_files_dispatch``.

    :return CheckUploadedFilesEquation: The decorated function.
    """
    check_uploaded_files_dispatch.append(checker)

    @wraps(checker)
    def decorator(file_list: List[UploadFile]):
        return checker(file_list)

    return decorator


@register_check_uploaded_files
def check_uploaded_files_extensions(file_list: List[UploadFile]):
    """
    Check the list of uploaded files to count the file extensions.

    Raise UploadedFilesValidationError if the business rules are not met.

    :param List[UploadFile] file_list: Upload file list.
    """
    extension_counter = Counter(get_suffix(f) for f in file_list)
    logger.debug(f"Checking uploaded files extensions: {extension_counter=}")
    with UploadedFilesValidationError.check_expressions("Invalid file extensions") as check:
        check(
            extension_counter[".py"] == 1,
            "missing one application source file (.py)",
        )
        check(
            0 <= extension_counter[".yaml"] <= 1,
            "missing one (optional) application config file (.yaml)",
        )
        check(
            extension_counter[".j2"] + extension_counter[".jinja2"] >= 1,
            "missing one or more template files (.j2 and/or .jinja2)",
        )


@register_check_uploaded_files
def check_uploaded_files_content_type(file_list: List[UploadFile]):
    """
    Check the list of uploaded files to confirm they are specified as plain text.

    Raise UploadedFilesValidationError if the business rules are not met.

    :param List[UploadFile] file_list: Upload file list.
    """
    logger.debug("Checking uploaded files content types")
    expected = "text/plain"
    extension_list = ", ".join(syntax_validation_dispatch.keys())
    UploadedFilesValidationError.require_condition(
        all(f.content_type == expected for f in file_list if get_suffix(f) in syntax_validation_dispatch),
        f"The content of the files {extension_list} is expected to be {expected}",
    )


@register_check_uploaded_files
def check_uploaded_files_syntax(file_list: List[UploadFile]):
    """
    Check the list of uploaded files to verify their syntax.

    Raise UploadedFilesValidationError if the business rules are not met.

    :param List[UploadFile] file_list: Upload file list.
    """
    logger.debug("Checking uploaded files, validating source code")

    with UploadedFilesValidationError.check_expressions("Invalid syntax on uploaded file(s)") as check:
        for f in file_list:
            suffix = get_suffix(f)
            if suffix not in syntax_validation_dispatch:
                continue
            check(syntax_validation_dispatch[suffix](f.file.read()), f.filename)
            f.file.seek(0)


@register_check_uploaded_files
def check_uploaded_files_yaml_is_parsable(file_list: List[UploadFile]):
    """
    Check the list of uploaded files to verify if the yaml file agrees with the expected schema.

    Raise UploadedFilesValidationError if the business rules are not met.

    :param List[UploadFile] file_list: Upload file list.
    """
    logger.debug("Checking uploaded files, parsing yaml file")
    with UploadedFilesValidationError.handle_errors(
        "Not possible to get the configuration from the yaml file. "
        "Please, verify is all the required fields were provided."
    ):
        for f in file_list:
            if f.filename.endswith(".yaml"):
                config = yaml_safe_load(f.file.read())
                f.file.seek(0)
                ApplicationConfig(**config)


SyntaxValidationEquation = Callable[[Union[str, bytes]], bool]
"""Type alias describing the function signature used to validate file syntax."""

syntax_validation_dispatch: Dict[str, SyntaxValidationEquation] = {}
"""Dictionary mapping file extensions to the function used to validate their syntax."""


def register_syntax_validator(*file_extensions: str):
    """
    Use this decorator to register file syntax validation functions.

    It creates a new entry on ``validation_dispatch``, mapping the equation to
    the file extensions that are provided as arguments.

    Raise ValueError if the provided file extensions do not start with a dot.

    :return ValidationEquation: The decorated function.
    """

    def decorator(validator):
        for extension in file_extensions:
            require_condition(
                extension.startswith("."),
                f"File extensions should start with a dot, got the value: {extension}",
                raise_exc_class=ValueError,
            )
            syntax_validation_dispatch[extension] = validator

        @wraps(validator)
        def wrapper(source_code):
            return validator(source_code)

        return wrapper

    return decorator


@register_syntax_validator(".py")
def is_valid_python_file(source_code: Union[str, bytes]) -> bool:
    """
    Check if a given Python source code is valid by parsing it into an AST node.

    :param Union[str, bytes] source_code: Python source code.
    :return bool: Boolean indicating if the source code is valid or not.
    """
    try:
        ast_parse(source_code)
    except Exception:
        return False
    return True


@register_syntax_validator(".yaml")
def is_valid_yaml_file(yaml_file: Union[str, bytes]) -> bool:
    """
    Check if a given YAML file is valid by parsing it with yaml.safe_load.

    :param Union[str, bytes] yaml_file: YAML file.
    :return bool: Boolean indicating if the file is valid or not.
    """
    try:
        yaml_safe_load(yaml_file)
    except Exception:
        return False
    return True


@register_syntax_validator(".j2", ".jinja2")
def is_valid_jinja2_template(template: Union[str, bytes]) -> bool:
    """
    Check if a given jinja2 template is valid by creating a Template object and trying to render it.

    :param str template: Jinja2 template.
    :return bool: Boolean indicating if the template is valid or not.
    """
    if isinstance(template, bytes):
        _template = template.decode("utf-8")
    else:
        _template = template

    try:
        Template(_template).render(data={})
    except Exception:
        return False
    return True



"""
Router for the JobScript resource.
"""
import json
from pathlib import PurePath
from typing import List, Optional

from armasec import TokenPayload
from fastapi import APIRouter, Depends, HTTPException, Query, status
from jinja2 import Template
from loguru import logger
from yaml import safe_load

from jobbergate_api.apps.applications.models import applications_table
from jobbergate_api.apps.applications.schemas import ApplicationConfig, ApplicationResponse
from jobbergate_api.apps.job_scripts.models import job_scripts_table, searchable_fields, sortable_fields
from jobbergate_api.apps.job_scripts.schemas import (
    JobScriptCreateRequest,
    JobScriptPartialResponse,
    JobScriptResponse,
    JobScriptUpdateRequest,
)
from jobbergate_api.apps.permissions import Permissions
from jobbergate_api.file_validation import UploadedFilesValidationError
from jobbergate_api.pagination import Pagination, ok_response, package_response
from jobbergate_api.s3_manager import get_application_files_from_s3, s3man_jobscripts
from jobbergate_api.security import IdentityClaims, guard
from jobbergate_api.storage import (
    INTEGRITY_CHECK_EXCEPTIONS,
    database,
    render_sql,
    search_clause,
    sort_clause,
)

router = APIRouter()


def inject_sbatch_params(job_script_data_as_string: str, sbatch_params: List[str]) -> str:
    """
    Inject sbatch params into job script.

    Given the job script as job_script_data_as_string, inject the sbatch params in the correct location.
    """
    logger.debug("Preparing to inject sbatch params into job script")

    if not sbatch_params:
        logger.warning("Sbatch param list is empty")
        return job_script_data_as_string

    first_sbatch_index = job_script_data_as_string.find("#SBATCH")
    string_slice = job_script_data_as_string[first_sbatch_index:]
    line_end = string_slice.find("\n") + first_sbatch_index + 1

    inner_string = ""
    for parameter in sbatch_params:
        inner_string += "#SBATCH " + parameter + "\\n"

    new_job_script_data_as_string = (
        job_script_data_as_string[:line_end] + inner_string + job_script_data_as_string[line_end:]
    )

    logger.debug("Done injecting sbatch params into job script")
    return new_job_script_data_as_string


def render_template(template_files, param_dict_flat):
    """
    Render the templates as strings using jinja2.
    """
    logger.debug("Rendering the templates as strings using jinja2")
    for key, value in template_files.items():
        template = Template(value)
        rendered_js = template.render(data=param_dict_flat)
        template_files[key] = rendered_js
    job_script_data_as_string = json.dumps(template_files)
    return job_script_data_as_string


def build_job_script_data_as_string(application_id: int, application_config: ApplicationConfig) -> str:
    """
    Return the job_script_data_as_string from the S3 application and the templates.
    """
    logger.debug("Building the job script file from the S3 application and the templates")

    application_files = get_application_files_from_s3(application_id)

    default_template_path = PurePath(application_config.jobbergate_config.default_template)
    default_template_name = default_template_path.name

    with UploadedFilesValidationError.check_expressions("One or more application files are missing") as check:
        check(
            application_files.source_file,
            f"Application file was not found for {application_id=}",
        )
        check(
            application_files.templates,
            f"No template was found for {application_id=}",
        )
        check(
            default_template_name in application_files.templates,
            f"{default_template_name=} was found for {application_id=}",
        )

    # rename default template
    default_template_file = application_files.templates.pop(default_template_name)
    application_files.templates["application.sh"] = default_template_file

    param_dict_flat = {}
    for (key, value) in application_config.dict().items():
        if isinstance(value, dict):
            for nest_key, nest_value in value.items():
                param_dict_flat[nest_key] = nest_value
        else:
            param_dict_flat[key] = value

    job_script_data_as_string = render_template(application_files.templates, param_dict_flat)

    logger.debug("Done building the job script file")
    return job_script_data_as_string


@router.post(
    "/job-scripts",
    status_code=status.HTTP_201_CREATED,
    response_model=JobScriptResponse,
    description="Endpoint for job_script creation",
)
async def job_script_create(
    job_script: JobScriptCreateRequest,
    token_payload: TokenPayload = Depends(guard.lockdown(Permissions.JOB_SCRIPTS_EDIT)),
):
    """
    Create a new job script.

    Make a post request to this endpoint with the required values to create a new job script.
    """
    logger.debug(f"Creating {job_script=}")

    select_query = applications_table.select().where(applications_table.c.id == job_script.application_id)
    logger.trace(f"select_query = {render_sql(select_query)}")

    raw_application = await database.fetch_one(select_query)

    if not raw_application:
        message = f"Application with id={job_script.application_id} not found."
        logger.warning(message)
        raise HTTPException(status_code=status.HTTP_404_NOT_FOUND, detail=message)

    application = ApplicationResponse.parse_obj(raw_application)

    if application.application_uploaded is False:
        raise HTTPException(
            status_code=status.HTTP_422_UNPROCESSABLE_ENTITY,
            detail=f"Application with id={job_script.application_id} was not uploaded.",
        )

    identity_claims = IdentityClaims.from_token_payload(token_payload)

    create_dict = dict(
        **{k: v for (k, v) in job_script.dict(exclude_unset=True).items() if k != "param_dict"},
        job_script_owner_email=identity_claims.email,
    )

    # Use application_config from the application as a baseline of defaults
    logger.debug(f"APP CONFIG: {application.application_config}")
    param_dict = safe_load(application.application_config)

    # User supplied param dict is optional and may override defaults
    param_dict.update(**job_script.param_dict)

    application_config = ApplicationConfig(**param_dict)

    try:
        job_script_data_as_string = build_job_script_data_as_string(application.id, application_config)
    except UploadedFilesValidationError as e:
        message = str(e)
        logger.error(message)
        raise HTTPException(status_code=status.HTTP_404_NOT_FOUND, detail=message)

    sbatch_params = create_dict.pop("sbatch_params", [])
    job_script_data_as_string = inject_sbatch_params(job_script_data_as_string, sbatch_params)

    logger.debug("Inserting job_script")

    async with database.transaction():

        try:
            insert_query = job_scripts_table.insert().returning(job_scripts_table)
            logger.trace(f"insert_query = {render_sql(insert_query)}")
            job_script_data = await database.fetch_one(query=insert_query, values=create_dict)

            if job_script_data is None:
                message = "An error occurred when inserting the JobScript at the database."
                logger.error(message)
                raise HTTPException(
                    status_code=status.HTTP_400_BAD_REQUEST,
                    detail=message,
                )

            s3man_jobscripts[job_script_data["id"]] = job_script_data_as_string

        except INTEGRITY_CHECK_EXCEPTIONS as e:
            logger.error(f"Reverting database transaction: {str(e)}")
            raise HTTPException(status_code=status.HTTP_422_UNPROCESSABLE_ENTITY, detail=str(e))

        except KeyError as e:
            logger.error(f"Reverting database transaction: {str(e)}")
            raise HTTPException(status_code=status.HTTP_422_UNPROCESSABLE_ENTITY, detail=str(e))

    logger.debug(f"Job-script created: {job_script_data}")
    job_script_response = dict(
        **job_script_data,
        job_script_data_as_string=job_script_data_as_string,
    )
    return job_script_response


@router.get(
    "/job-scripts/{job_script_id}",
    description="Endpoint to get a job_script",
    response_model=JobScriptResponse,
    dependencies=[Depends(guard.lockdown(Permissions.JOB_SCRIPTS_VIEW))],
)
async def job_script_get(job_script_id: int = Query(...)):
    """
    Return the job_script given its id.
    """
    logger.debug(f"Getting {job_script_id=}")

    query = job_scripts_table.select().where(job_scripts_table.c.id == job_script_id)
    logger.trace(f"get_query = {render_sql(query)}")
    job_script = await database.fetch_one(query)

    if not job_script:
        message = f"JobScript with id={job_script_id} not found."
        logger.warning(message)
        raise HTTPException(
            status_code=status.HTTP_404_NOT_FOUND,
            detail=message,
        )

    job_script_response = dict(job_script)

    try:
        job_script_response["job_script_data_as_string"] = s3man_jobscripts[job_script_id]
    except KeyError:
        message = f"JobScript file not found for id={job_script_id}."
        logger.warning(message)
        raise HTTPException(
            status_code=status.HTTP_404_NOT_FOUND,
            detail=message,
        )

    logger.debug(f"Job-script data: {job_script_response=}")

    return job_script_response


@router.get(
    "/job-scripts",
    description="Endpoint to list job_scripts",
    responses=ok_response(JobScriptPartialResponse),
)
async def job_script_list(
    pagination: Pagination = Depends(),
    all: Optional[bool] = Query(False),
    search: Optional[str] = Query(None),
    sort_field: Optional[str] = Query(None),
    sort_ascending: bool = Query(True),
    token_payload: TokenPayload = Depends(guard.lockdown(Permissions.JOB_SCRIPTS_VIEW)),
):
    """
    List job_scripts for the authenticated user.

    Note::

       Use responses instead of response_model to skip a second round of validation and serialization. This
       is already happening in the ``package_response`` method. So, we uses ``responses`` so that FastAPI
       can generate the correct OpenAPI spec but not post-process the response.
    """
    logger.debug("Preparing to list job-scripts")

    query = job_scripts_table.select()
    identity_claims = IdentityClaims.from_token_payload(token_payload)
    if not all:
        query = query.where(job_scripts_table.c.job_script_owner_email == identity_claims.email)
    if search is not None:
        query = query.where(search_clause(search, searchable_fields))
    if sort_field is not None:
        query = query.order_by(sort_clause(sort_field, sortable_fields, sort_ascending))

    logger.trace(f"Query = {render_sql(query)}")
    return await package_response(JobScriptPartialResponse, query, pagination)


@router.delete(
    "/job-scripts/{job_script_id}",
    status_code=status.HTTP_204_NO_CONTENT,
    description="Endpoint to delete job script",
    dependencies=[Depends(guard.lockdown(Permissions.JOB_SCRIPTS_EDIT))],
)
async def job_script_delete(job_script_id: int = Query(..., description="id of the job script to delete")):
    """
    Delete job_script given its id.
    """
    logger.debug(f"Preparing to delete {job_script_id=}")
    where_stmt = job_scripts_table.c.id == job_script_id

    get_query = job_scripts_table.select().where(where_stmt)
    logger.trace(f"get_query = {render_sql(get_query)}")

    raw_job_script = await database.fetch_one(get_query)
    if not raw_job_script:

        message = f"JobScript with id={job_script_id} not found."
        logger.warning(message)
        raise HTTPException(
            status_code=status.HTTP_404_NOT_FOUND,
            detail=message,
        )

    delete_query = job_scripts_table.delete().where(where_stmt)
    logger.trace(f"delete_query = {render_sql(delete_query)}")
    await database.execute(delete_query)

    try:
        del s3man_jobscripts[job_script_id]
    except KeyError:
        # There is no need to raise an error if we try to delete a file that does not exist
        logger.warning(f"Tried to delete {job_script_id=}, but it was not found on S3.")


@router.put(
    "/job-scripts/{job_script_id}",
    status_code=status.HTTP_200_OK,
    description="Endpoint to update a job_script given the id",
    response_model=JobScriptResponse,
    dependencies=[Depends(guard.lockdown(Permissions.JOB_SCRIPTS_EDIT))],
)
async def job_script_update(job_script_id: int, job_script: JobScriptUpdateRequest):
    """
    Update a job_script given its id.
    """
    logger.debug(f"Updating {job_script_id=}")

    update_query = (
        job_scripts_table.update()
        .where(job_scripts_table.c.id == job_script_id)
        .values(
            job_script.dict(exclude_unset=True, exclude={"job_script_data_as_string"}),
        )
        .returning(job_scripts_table)
    )
    logger.trace(f"update_query = {render_sql(update_query)}")

    async with database.transaction():

        try:
            result = await database.fetch_one(update_query)
        except INTEGRITY_CHECK_EXCEPTIONS as e:
            logger.error(f"Reverting database transaction: {str(e)}")
            raise HTTPException(status_code=status.HTTP_409_CONFLICT, detail=str(e))

        if result is None:
            message = f"JobScript with id={job_script_id} not found."
            logger.warning(message)
            raise HTTPException(
                status_code=status.HTTP_404_NOT_FOUND,
                detail=message,
            )

        job_script_response = dict(result)

        try:
            if job_script.job_script_data_as_string:
                s3man_jobscripts[job_script_id] = job_script.job_script_data_as_string
                job_script_response["job_script_data_as_string"] = job_script.job_script_data_as_string
            else:
                job_script_response["job_script_data_as_string"] = s3man_jobscripts[job_script_id]
        except KeyError as e:
            logger.error(f"Reverting database transaction: {str(e)}")
            raise HTTPException(
                status_code=status.HTTP_404_NOT_FOUND,
                detail=f"File for JobScript with id={job_script_id} not found in S3.",
            )

    return job_script_response


def include_router(app):
    """
    Include the router for this module in the app.
    """
    app.include_router(router)



"""
Test the email notification system at Jobbergate.
"""
from unittest import mock

import pytest
from fastapi import HTTPException
from sendgrid import SendGridAPIClient
from sendgrid.helpers.mail import Mail

from jobbergate_api.email_notification import EmailManager, EmailNotificationError, notify_submission_rejected


class TestEmailManager:
    """
    A suite of tests for the email manager.
    """

    @pytest.fixture(scope="class")
    def email_client(self):
        """
        Build a dummy email client for tests.
        """
        return SendGridAPIClient("dummy-api-key")

    @pytest.fixture(scope="class")
    def email_manager(self, email_client):
        """
        Build a dummy email manager for tests.
        """
        return EmailManager(email_client, "me@pytesting.com")

    @pytest.fixture(scope="class")
    def email_options(self):
        """
        Return a dictionary with the parameters used to send an email using the manager.
        """
        return dict(
            to_emails="you@pytesting.com",
            subject="TEST",
            plain_text_content="The content!",
        )

    def test_build_message__success(self, email_manager, email_options):
        """
        Test that a message was usefully built and returned using the provided parameters.
        """
        message = email_manager._build_message(**email_options)
        assert isinstance(message, Mail)

    def test_build_message__error_when_from_email_is_empty(self, email_client, email_options):
        """
        Test if build_message raises EmailNotificationError when the parameter from_email is empty.
        """
        email_manager = EmailManager(email_client, "")
        with pytest.raises(EmailNotificationError, match="The value from_email is empty"):
            email_manager._build_message(**email_options)

    def test_build_message__error_at_mail_arguments(self, email_manager, email_options):
        """
        Test if build_message raises EmailNotificationError.

        In this case, when some invalid parameters are provided to the Mail constructor.
        """
        dummy_options = dict(**email_options, foo="foo", bar="bar")
        with pytest.raises(EmailNotificationError, match="Error while creating the message"):
            email_manager._build_message(**dummy_options)

    def test_send_email__success(self, email_manager, email_options):
        """
        Test that the email manager can send an email successfully using the SendGrid API.
        """
        with mock.patch.object(email_manager.email_client, "send") as mocked:
            email_manager.send_email(**email_options)
        mocked.assert_called_once()

    def test_send_email__fail(self, email_manager, email_options):
        """
        Test if the email manager raises EmailNotificationError.

        In this case, when it can not connect to the SendGrid API.
        """
        with pytest.raises(EmailNotificationError, match="Error while sending email"):
            with mock.patch.object(
                email_manager.email_client,
                "send",
                side_effect=HTTPException(status_code=401, detail="Unauthorized"),
            ) as mocked:
                email_manager.send_email(**email_options)
        mocked.assert_called_once()

    def test_send_email__skip_on_failure(self, email_manager, email_options):
        """
        Test if the email manager can skip on failure when it can not connect to the SendGrid API.
        """
        with mock.patch.object(
            email_manager.email_client,
            "send",
            side_effect=HTTPException(status_code=401, detail="Unauthorized"),
        ) as mocked:
            email_manager.send_email(skip_on_failure=True, **email_options)
        mocked.assert_called_once()


def test_notify_submission_rejected():
    """
    Test if an email is sent to inform that a job submission has been rejected.
    """
    with mock.patch("jobbergate_api.email_notification.email_manager.email_client.send") as mocked:
        notify_submission_rejected(
            job_submission_id=0,
            report_message="something went wrong!",
            to_emails=["support@pytesting.com", "someone@pytesting.com"],
        )
    mocked.assert_called_once()



"""
Test s3 manager.
"""

import contextlib
from unittest.mock import patch

import pytest
from fastapi import UploadFile
from file_storehouse import FileManager, FileManagerReadOnly  # type: ignore

from jobbergate_api.s3_manager import (
    ApplicationFiles,
    application_template_manager_factory,
    delete_application_files_from_s3,
    get_application_files_from_s3,
    s3man_applications_source_files,
    s3man_jobscripts,
    write_application_files_to_s3,
)


@pytest.mark.parametrize(
    "s3_manager, template",
    [
        (s3man_applications_source_files, "applications/{}/jobbergate.py"),
        (s3man_jobscripts, "job-scripts/{}/jobbergate.txt"),
    ],
)
@pytest.mark.parametrize("id", [0, 1, 2, 10, 100, 9999])
class TestS3ManagerKeyIdTwoWayMapping:
    """
    Test the conversions from id number to S3 key and vice versa.
    """

    @pytest.mark.parametrize("input_type", [int, str])
    def test_s3_manager__get_key_from_id_str(self, s3_manager, template, id, input_type):
        """
        Test the conversions from id number to S3 key.

        Notice both int and str are valid types for id and are tested.
        """
        key = template.format(id)
        assert s3_manager._get_engine_key(input_type(id)) == key

    def test_s3_manager__get_app_id_from_key(self, s3_manager, template, id):
        """
        Test the conversions from S3 key to id number.
        """
        key = template.format(id)
        assert s3_manager._get_dict_key(key) == id


@pytest.mark.parametrize(
    "ManagerClass, is_read_only",
    [
        (FileManager, False),
        (FileManagerReadOnly, True),
    ],
)
def test_application_template_manager_factory_returning_type(ManagerClass, is_read_only):
    """
    Test if the object returned by application_template_manager_factory is the expected.
    """
    assert isinstance(
        application_template_manager_factory(0, is_read_only=is_read_only),
        ManagerClass,
    )


@pytest.fixture
def make_uploaded_files():
    """
    Provide a fixture to use as a context manager that builds a list of uploaded files.
    """

    @contextlib.contextmanager
    def _helper(*file_paths):
        """
        Context manager that opens the file(s) and yields the ``files`` param from it.
        """
        with contextlib.ExitStack() as stack:
            yield [
                UploadFile(path.name, stack.enter_context(open(path)), "text/plain") for path in file_paths
            ]

    return _helper


@pytest.fixture(scope="function")
def mocked_application_template_manager():
    """
    Mock the template manager factory in order to make it return an empty dictionary.

    :yield dict: The dictionary representing the file manager.
    """
    mock_result_as_dict = {}
    with patch(
        "jobbergate_api.s3_manager.application_template_manager_factory",
        wraps=lambda *args, **kwargs: mock_result_as_dict,
    ):
        yield mock_result_as_dict
    mock_result_as_dict.clear()


@pytest.fixture(scope="function")
def mocked_applications_source_file_manager():
    """
    Mock the application source file manager in order to make it return an empty dictionary.

    :yield dict: The dictionary representing the file manager.
    """
    mock_result_as_dict = {}
    with patch("jobbergate_api.s3_manager.s3man_applications_source_files", mock_result_as_dict):
        yield mock_result_as_dict
    mock_result_as_dict.clear()


def test_write_application_files_to_s3(
    mocked_application_template_manager,
    mocked_applications_source_file_manager,
    make_dummy_file,
    make_uploaded_files,
    dummy_application_source_file,
    dummy_template,
    dummy_application_config,
):
    """
    Test if the applications files are written to S3 as expected.

    To do so, a list of uploaded files is created, than  we verify if each file
    went to the correct file manager.
    """
    application_id = 1

    with make_uploaded_files(
        make_dummy_file("jobbergate.py", content=dummy_application_source_file),
        make_dummy_file("template-1.j2", content=dummy_template),
        make_dummy_file("template-2.jinja2", content=dummy_template),
        make_dummy_file("jobbergate.yaml", content=dummy_application_config),
    ) as uploaded_files:
        write_application_files_to_s3(application_id, uploaded_files, remove_previous_files=True)

    expected_templates = {"template-1.j2", "template-2.jinja2"}
    assert mocked_application_template_manager.keys() == expected_templates

    expected_source_files = {application_id}
    assert mocked_applications_source_file_manager.keys() == expected_source_files


def test_get_application_files_from_s3(
    mocked_application_template_manager,
    mocked_applications_source_file_manager,
    dummy_application_source_file,
    dummy_template,
):
    """
    Test if the application files are loaded from S3 as expected.

    To do so, dummy test files are added to the file managers and than loaded.
    """
    application_id = 1

    mocked_applications_source_file_manager[application_id] = dummy_application_source_file
    mocked_application_template_manager["template-1.j2"] = dummy_template
    mocked_application_template_manager["template-2.jinja2"] = dummy_template

    application_files = get_application_files_from_s3(application_id)

    assert isinstance(application_files, ApplicationFiles)

    assert application_files.source_file == dummy_application_source_file

    assert application_files.templates == mocked_application_template_manager


def test_delete_application_files_from_s3(
    mocked_application_template_manager,
    mocked_applications_source_file_manager,
):
    """
    Test if the applications files are deleted from S3 as expected.

    To do so, test files are added to both file managers. Once one application id
    is deleted, all templated files related to it are deleted as well as its source
    file. Source files from any other id are expected to be preserved.
    """
    FILES_CONTENT = "test-file"

    mocked_applications_source_file_manager.update((i, FILES_CONTENT) for i in range(3))
    mocked_application_template_manager.update((f"{i}.j2", FILES_CONTENT) for i in range(3))

    delete_application_files_from_s3(application_id=1)

    assert mocked_applications_source_file_manager.keys() == {0, 2}
    assert len(mocked_application_template_manager) == 0



"""
Main file to startup the fastapi server.
"""
import sys
import typing

import asyncpg
import sentry_sdk
from fastapi import FastAPI, Response, status
from loguru import logger
from sentry_sdk.integrations.asgi import SentryAsgiMiddleware
from starlette.middleware.cors import CORSMiddleware

from jobbergate_api.apps.applications.routers import router as applications_router
from jobbergate_api.apps.job_scripts.routers import router as job_scripts_router
from jobbergate_api.apps.job_submissions.routers import router as job_submissions_router
from jobbergate_api.config import settings
from jobbergate_api.storage import database, handle_fk_error

subapp = FastAPI()
subapp.add_middleware(
    CORSMiddleware,
    allow_origins=["*"],
    allow_credentials=True,
    allow_methods=["*"],
    allow_headers=["*"],
)

if settings.SENTRY_DSN and settings.DEPLOY_ENV.lower() != "test":
    logger.info("Initializing Sentry")
    sentry_sdk.init(
        dsn=settings.SENTRY_DSN,
        sample_rate=typing.cast(float, settings.SENTRY_SAMPLE_RATE),  # The cast silences mypy
        environment=settings.DEPLOY_ENV,
    )
    subapp.add_middleware(SentryAsgiMiddleware)
else:
    logger.info("Skipping Sentry")

subapp.include_router(applications_router)
subapp.include_router(job_scripts_router)
subapp.include_router(job_submissions_router)
subapp.exception_handler(asyncpg.exceptions.ForeignKeyViolationError)(handle_fk_error)


@subapp.get(
    "/health",
    status_code=status.HTTP_204_NO_CONTENT,
    responses={204: {"description": "API is healthy"}},
)
async def health_check():
    """
    Provide a health-check endpoint for the app.
    """
    return Response(status_code=status.HTTP_204_NO_CONTENT)


app = FastAPI()
app.mount("/jobbergate", subapp)


@app.on_event("startup")
def init_logger():
    """
    Initialize logging.
    """
    logger.remove()
    logger.add(sys.stderr, level=settings.LOG_LEVEL)
    logger.info(f"Logging configured 📝 Level: {settings.LOG_LEVEL}")


@app.on_event("startup")
async def init_database():
    """
    Connect the database; create it if necessary.
    """
    logger.debug("Initializing database")
    await database.connect()


@app.on_event("shutdown")
async def disconnect_database():
    """
    Disconnect the database.
    """
    logger.debug("Disconnecting database")
    await database.disconnect()



"""
Provide a convenience class for managing calls to S3.
"""
import typing
from dataclasses import dataclass
from functools import lru_cache
from pathlib import PurePath

from fastapi import UploadFile
from file_storehouse import FileManager, FileManagerReadOnly, client
from file_storehouse.engine import EngineS3
from file_storehouse.key_mapping import KeyMappingNumeratedFolder, KeyMappingRaw
from file_storehouse.transformation import TransformationCodecs
from loguru import logger

from jobbergate_api.config import settings
from jobbergate_api.file_validation import perform_all_checks_on_uploaded_files


@lru_cache
def application_template_manager_factory(
    application_id: int, is_read_only: bool = True
) -> typing.Union[FileManager, FileManagerReadOnly]:
    """
    Build a manager object for application template files.

    :param int application_id: Application's id
    :param bool is_read_only: If the manager is read only or not, defaults to True
    :return typing.Union[FileManager, FileManagerReadOnly]: File manager
    """
    Manager = FileManagerReadOnly if is_read_only else FileManager
    return Manager(
        engine=EngineS3(
            s3_client,
            settings.S3_BUCKET_NAME,
            f"applications/{application_id}/templates/",
        ),
        transformation_list=[TransformationCodecs("utf-8")],
        key_mapping=KeyMappingRaw(),
    )


def write_application_files_to_s3(
    application_id: int,
    upload_files: typing.List[UploadFile],
    *,
    remove_previous_files: bool = False,
):
    """
    Write the list of uploaded application files to S3, fist checking them for consistency.

    :param int application_id: Application identification number
    :param typing.List[UploadFile] upload_files: Uploaded files
    :param bool remove_previous_files: Delete old files before writing the new ones
    """
    logger.debug(f"Writing the list of uploaded files to S3: {application_id=}")

    perform_all_checks_on_uploaded_files(upload_files)

    if remove_previous_files:
        delete_application_files_from_s3(application_id)

    templates_manager = typing.cast(
        FileManager,
        application_template_manager_factory(application_id, is_read_only=False),
    )

    for upload in upload_files:
        if upload.filename.endswith(".py"):
            s3man_applications_source_files[application_id] = upload.file
        elif upload.filename.endswith((".j2", ".jinja2")):
            filename = PurePath(upload.filename).name
            templates_manager[filename] = upload.file


@dataclass
class ApplicationFiles:
    """
    Dataclass containing application files.
    """

    templates: typing.Dict[str, str]
    source_file: str


def get_application_files_from_s3(application_id: int) -> ApplicationFiles:
    """
    Read the application files from S3.

    :param int application_id: Application identification number
    :return ApplicationFiles: Application files
    """
    logger.debug(f"Getting application files from S3: {application_id=}")

    templates_manager = application_template_manager_factory(application_id, True)

    return ApplicationFiles(
        templates=dict(templates_manager), source_file=s3man_applications_source_files.get(application_id, "")
    )


def delete_application_files_from_s3(application_id: int):
    """
    Delete the files associated to a given id from S3.

    :param int application_id: Application identification number
    """
    logger.debug(f"Deleting from S3 the files associated to {application_id=}")

    templates_manager = typing.cast(
        FileManager, application_template_manager_factory(application_id, is_read_only=False)
    )

    templates_manager.clear()
    try:
        del s3man_applications_source_files[application_id]
    except KeyError:
        logger.warning(
            f"Tried to delete the source code for {application_id=}, but it was not found on S3",
        )


s3_client = client(
    "s3",
    endpoint_url=settings.S3_ENDPOINT_URL,
)

s3man_applications_source_files = FileManager(
    engine=EngineS3(s3_client, settings.S3_BUCKET_NAME, "applications"),
    transformation_list=[TransformationCodecs("utf-8")],
    key_mapping=KeyMappingNumeratedFolder("jobbergate.py"),
)

s3man_jobscripts = FileManager(
    engine=EngineS3(s3_client, settings.S3_BUCKET_NAME, "job-scripts"),
    transformation_list=[TransformationCodecs("utf-8")],
    key_mapping=KeyMappingNumeratedFolder("jobbergate.txt"),
)



"""
Database model for the JobSubmission resource.
"""
from sqlalchemy import DateTime, Enum, ForeignKey, Integer, String, Table
from sqlalchemy.sql import func
from sqlalchemy.sql.schema import Column

from jobbergate_api.apps.job_submissions.constants import JobSubmissionStatus
from jobbergate_api.metadata import metadata

job_submissions_table = Table(
    "job_submissions",
    metadata,
    Column("id", Integer, primary_key=True),
    Column("job_submission_name", String, nullable=False, index=True),
    Column("job_submission_description", String, default=""),
    Column("job_submission_owner_email", String, nullable=False, index=True),
    Column("job_script_id", ForeignKey("job_scripts.id"), nullable=False),
    Column("execution_directory", String),
    Column("slurm_job_id", Integer, default=None),
    Column("client_id", String, nullable=False, index=True),
    Column("status", Enum(JobSubmissionStatus), nullable=False, index=True),
    Column("report_message", String, nullable=True),
    Column("created_at", DateTime, nullable=False, default=func.now()),
    Column("updated_at", DateTime, nullable=False, default=func.now(), onupdate=func.now()),
)

searchable_fields = [
    job_submissions_table.c.job_submission_name,
    job_submissions_table.c.job_submission_description,
    job_submissions_table.c.job_submission_owner_email,
    job_submissions_table.c.client_id,
]

sortable_fields = [
    job_submissions_table.c.id,
    job_submissions_table.c.job_submission_name,
    job_submissions_table.c.job_submission_owner_email,
    job_submissions_table.c.job_script_id,
    job_submissions_table.c.slurm_job_id,
    job_submissions_table.c.client_id,
    job_submissions_table.c.created_at,
    job_submissions_table.c.updated_at,
]



"""
Tests for the /applications/ endpoint.
"""
import json
from unittest import mock

import asyncpg
import pytest
from fastapi import status

from jobbergate_api.apps.applications.models import applications_table
from jobbergate_api.apps.applications.schemas import ApplicationPartialResponse
from jobbergate_api.apps.permissions import Permissions
from jobbergate_api.s3_manager import ApplicationFiles
from jobbergate_api.storage import database, fetch_instance


@pytest.mark.asyncio
async def test_create_application(
    fill_application_data,
    client,
    inject_security_header,
    time_frame,
):
    """
    Test POST /applications/ correctly creates an application.

    This test proves that an application is successfully created via a POST request to the /applications/
    endpoint. We show this by asserting that the application is created in the database after the post
    request is made and the correct status code (201) is returned.
    """
    id_rows = await database.fetch_all("SELECT id FROM applications")
    assert len(id_rows) == 0

    inject_security_header("owner1@org.com", Permissions.APPLICATIONS_EDIT)
    with time_frame() as window:
        response = await client.post(
            "/jobbergate/applications/",
            json=fill_application_data(
                application_name="test-name",
                application_identifier="test-identifier",
            ),
        )

    assert response.status_code == status.HTTP_201_CREATED

    id_rows = await database.fetch_all("SELECT id FROM applications")
    assert len(id_rows) == 1

    application = ApplicationPartialResponse(**response.json())

    assert application.id == id_rows[0][0]
    assert application.application_name == "test-name"
    assert application.application_identifier == "test-identifier"
    assert application.application_owner_email == "owner1@org.com"
    assert application.application_description is None
    assert application.created_at in window
    assert application.updated_at in window


@pytest.mark.asyncio
async def test_create_application_bad_permission(
    application_data,
    client,
    inject_security_header,
):
    """
    Test that it is not possible to create application without proper permission.

    This test proves that is not possible to create an application without the proper permission.
    We show this by trying to create an application without an permission that allow "create" then assert
    that the application still does not exists in the database and that the correct status code (403) is
    returned.
    """
    count = await database.fetch_all("SELECT COUNT(*) FROM applications")
    assert count[0][0] == 0

    inject_security_header("owner1@org.com", "INVALID_PERMISSION")
    response = await client.post("/jobbergate/applications/", json=application_data)
    assert response.status_code == status.HTTP_403_FORBIDDEN

    count = await database.fetch_all("SELECT COUNT(*) FROM applications")
    assert count[0][0] == 0


@pytest.mark.asyncio
async def test_create_without_application_name(
    application_data,
    client,
    inject_security_header,
):
    """
    Test that is not possible to create an application without the required body fields.

    This test proves that is not possible to create an application without the name. We show this by
    trying to create an application without the application_name in the request then assert that the
    application still does not exists in the database and the correct status code (422) is returned.
    """
    inject_security_header("owner1@org.com", Permissions.APPLICATIONS_EDIT)
    response = await client.post(
        "/jobbergate/applications/",
        json={k: v for (k, v) in application_data.items() if k != "application_name"},
    )
    assert response.status_code == status.HTTP_422_UNPROCESSABLE_ENTITY

    count = await database.fetch_all("SELECT COUNT(*) FROM applications")
    assert count[0][0] == 0


@pytest.mark.asyncio
@mock.patch("jobbergate_api.apps.applications.routers.delete_application_files_from_s3")
async def test_delete_application_no_file_uploaded(
    mocked_application_deleter, client, application_data, inject_security_header
):
    """
    Test DELETE /applications/<id> correctly deletes an application.

    This test proves that an application is successfully deleted via a DELETE request to the
    /applications/<id> endpoint. We show this by asserting that the application no longer exists in the
    database after the delete request is made and the correct status code is returned.
    """
    inserted_id = await database.execute(
        query=applications_table.insert(),
        values=application_data,
    )
    count = await database.fetch_all("SELECT COUNT(*) FROM applications")
    assert count[0][0] == 1

    inject_security_header("owner1@org.com", Permissions.APPLICATIONS_EDIT)
    response = await client.delete(f"/jobbergate/applications/{inserted_id}")

    assert response.status_code == status.HTTP_204_NO_CONTENT
    count = await database.fetch_all("SELECT COUNT(*) FROM applications")
    assert count[0][0] == 0

    mocked_application_deleter.assert_called_once_with(inserted_id)


@pytest.mark.asyncio
@mock.patch("jobbergate_api.apps.applications.routers.delete_application_files_from_s3")
async def test_delete_application_with_uploaded_file(
    mocked_application_deleter, client, application_data, inject_security_header
):
    """
    Test DELETE /applications/<id> correctly deletes an application and it's file.

    This test proves that an application is successfully deleted via a DELETE request to the
    /applications/<id> endpoint. We show this by asserting that the application no longer exists in the
    database after the delete request is made, the correct status code is returned and the correct boto3
    method was called.
    """
    inserted_id = await database.execute(
        query=applications_table.insert(),
        values=application_data,
    )
    count = await database.fetch_all("SELECT COUNT(*) FROM applications")
    assert count[0][0] == 1

    inject_security_header("owner1@org.com", Permissions.APPLICATIONS_EDIT)
    response = await client.delete(f"/jobbergate/applications/{inserted_id}")

    assert response.status_code == status.HTTP_204_NO_CONTENT
    count = await database.fetch_all("SELECT COUNT(*) FROM applications")
    assert count[0][0] == 0

    mocked_application_deleter.assert_called_once_with(inserted_id)


@pytest.mark.asyncio
@mock.patch("jobbergate_api.apps.applications.routers.delete_application_files_from_s3")
async def test_delete_application_by_identifier(
    mocked_application_deleter, client, fill_application_data, inject_security_header
):
    """
    Test DELETE /applications?identifier=<identifier> correctly deletes an application and it's file.

    This test proves that an application is successfully deleted via a DELETE request to the
    /applications?identifier=<identifier> endpoint. We show this by asserting that the application no longer
    exists in the database after the delete request is made, the correct status code is returned and the
    correct boto3 method was called.
    """
    inserted_id = await database.execute(
        query=applications_table.insert(),
        values=fill_application_data(
            application_owner_email="owner1@org.com",
            application_identifier="test-identifier",
        ),
    )
    count = await database.fetch_all("SELECT COUNT(*) FROM applications")
    assert count[0][0] == 1

    inject_security_header("owner1@org.com", Permissions.APPLICATIONS_EDIT)
    response = await client.delete("/jobbergate/applications?identifier=test-identifier")

    assert response.status_code == status.HTTP_204_NO_CONTENT
    count = await database.fetch_all("SELECT COUNT(*) FROM applications")
    assert count[0][0] == 0

    mocked_application_deleter.assert_called_once_with(inserted_id)


@pytest.mark.asyncio
async def test_delete_application_bad_permission(
    client,
    application_data,
    inject_security_header,
):
    """
    Test that it is not possible to delete application without proper permission.

    This test proves that an application is not deleted via a DELETE request to the /applications/<id>
    endpoint. We show this by asserting that the application still exists in the database after the delete
    request is made and the correct status code is returned.
    """
    inserted_id = await database.execute(
        query=applications_table.insert(),
        values=application_data,
    )
    count = await database.fetch_all("SELECT COUNT(*) FROM applications")
    assert count[0][0] == 1

    inject_security_header("owner1@org.com", "INVALID_PERMISSION")
    response = await client.delete(f"/jobbergate/applications/{inserted_id}")
    assert response.status_code == status.HTTP_403_FORBIDDEN
    count = await database.fetch_all("SELECT COUNT(*) FROM applications")
    assert count[0][0] == 1


@pytest.mark.asyncio
async def test_delete_application_not_found(client, inject_security_header):
    """
    Test DELETE /applications/<id> the correct response code when the application doesn't exist.

    This test proves that DELETE /applications/<id> returns the correct response code (404)
    when the application id does not exist in the database. We show this by asserting that a 404 response
    code is returned for a request made with an application id that doesn't exist.
    """
    inject_security_header("owner1@org.com", Permissions.APPLICATIONS_EDIT)
    response = await client.delete("/jobbergate/applications/999")
    assert response.status_code == status.HTTP_404_NOT_FOUND


@pytest.mark.asyncio
async def test_delete_application__fk_error(
    client,
    application_data,
    inject_security_header,
):
    """
    Test DELETE /applications/<id> correctly returns a 409 when a foreign-key error occurs.

    Test that a helpful message when a delete is blocked by a foreign-key constraint.
    """
    inserted_id = await database.execute(
        query=applications_table.insert(),
        values=application_data,
    )
    count = await database.fetch_all("SELECT COUNT(*) FROM applications")
    assert count[0][0] == 1

    inject_security_header("owner1@org.com", Permissions.APPLICATIONS_EDIT)
    with mock.patch(
        "jobbergate_api.storage.database.execute",
        side_effect=asyncpg.exceptions.ForeignKeyViolationError(
            f"""
            update or delete on table "applications" violates foreign key constraint
            "job_scripts_application_id_fkey" on table "job_scripts"
            DETAIL:  Key (id)=({inserted_id}) is still referenced from table "job_scripts".
            """
        ),
    ):
        response = await client.delete(f"/jobbergate/applications/{inserted_id}")
    assert response.status_code == status.HTTP_409_CONFLICT
    error_data = json.loads(response.text)["detail"]
    assert error_data["message"] == "Delete failed due to foreign-key constraint"
    assert error_data["table"] == "job_scripts"
    assert error_data["pk_id"] == f"{inserted_id}"


@pytest.mark.asyncio
async def test_get_application_by_id__files_not_uploaded(
    client,
    fill_application_data,
    inject_security_header,
):
    """
    Test GET /applications/<id> when the application files were not uploaded.

    This test proves that GET /applications/<id> returns the correct application, owned by
    the user making the request. We show this by asserting that the application data
    returned in the response is equal to the application data that exists in the database
    for the given application id.
    """
    inserted_id = await database.execute(
        query=applications_table.insert(),
        values=fill_application_data(application_identifier="app1"),
    )
    await database.execute(
        query=applications_table.insert(),
        values=fill_application_data(application_identifier="app2"),
    )
    count = await database.fetch_all("SELECT COUNT(*) FROM applications")
    assert count[0][0] == 2

    inject_security_header("owner1@org.com", Permissions.APPLICATIONS_VIEW)
    response = await client.get(f"/jobbergate/applications/{inserted_id}")
    assert response.status_code == status.HTTP_200_OK

    data = response.json()
    assert data["id"] == inserted_id
    assert data["application_identifier"] == "app1"
    assert data["application_uploaded"] is False
    assert data["application_config"] is None
    assert data["application_source_file"] is None
    assert data["application_templates"] is None


@pytest.mark.asyncio
@mock.patch("jobbergate_api.apps.applications.routers.get_application_files_from_s3")
async def test_get_application_by_id__files_uploaded(
    mocked_get_application_files_from_s3,
    client,
    fill_application_data,
    inject_security_header,
    dummy_application_config,
    dummy_application_source_file,
    dummy_template,
):
    """
    Test GET /applications/<id> when the application files were uploaded.

    This test proves that GET /applications/<id> returns the correct application, owned by
    the user making the request. We show this by asserting that the application data
    returned in the response is equal to the application data that exists in the database
    for the given application id, and the application files are recovered from S3.
    """
    inserted_id = await database.execute(
        query=applications_table.insert(),
        values=fill_application_data(
            application_identifier="app1",
            application_uploaded=True,
            application_config=dummy_application_config,
        ),
    )
    await database.execute(
        query=applications_table.insert(),
        values=fill_application_data(application_identifier="app2"),
    )

    mocked_get_application_files_from_s3.return_value = ApplicationFiles(
        templates={"test_job_script.sh": dummy_template},
        source_file=dummy_application_source_file,
    )

    count = await database.fetch_all("SELECT COUNT(*) FROM applications")
    assert count[0][0] == 2

    inject_security_header("owner1@org.com", Permissions.APPLICATIONS_VIEW)
    response = await client.get(f"/jobbergate/applications/{inserted_id}")
    assert response.status_code == status.HTTP_200_OK

    data = response.json()
    assert data["id"] == inserted_id
    assert data["application_identifier"] == "app1"
    assert data["application_uploaded"] is True
    assert data["application_config"] == dummy_application_config
    assert data["application_source_file"] == dummy_application_source_file
    assert data["application_templates"] == {"test_job_script.sh": dummy_template}


@pytest.mark.asyncio
async def test_get_application_by_id_invalid(client, inject_security_header):
    """
    Test the correct response code is returned when an application does not exist.

    This test proves that GET /application/<id> returns the correct response code when the
    requested application does not exist. We show this by asserting that the status code
    returned is what we would expect given the application requested doesn't exist (404).
    """
    inject_security_header("owner1@org.com", Permissions.APPLICATIONS_VIEW)
    response = await client.get("/jobbergate/applications/10")
    assert response.status_code == status.HTTP_404_NOT_FOUND


@pytest.mark.asyncio
async def test_get_application_by_id_bad_permission(
    client,
    application_data,
    inject_security_header,
):
    """
    Test that it is not possible to get application without proper permission.

    This test proves that GET /application/<id> returns the correct response code when the
    user don't have the proper permission. We show this by asserting that the status code
    returned is what we would expect (403).
    """
    inserted_id = await database.execute(
        query=applications_table.insert(),
        values=application_data,
    )

    inject_security_header("owner1@org.com", "INVALID_PERMISSION")
    response = await client.get(f"/jobbergate/applications/{inserted_id}")
    assert response.status_code == status.HTTP_403_FORBIDDEN


@pytest.mark.asyncio
async def test_get_applications__no_params(
    client,
    fill_all_application_data,
    inject_security_header,
):
    """
    Test GET /applications returns only applications owned by the user making the request.

    This test proves that GET /applications returns the correct applications for the user making
    the request. We show this by asserting that the applications returned in the response are
    only applications owned by the user making the request.
    """
    await database.execute_many(
        applications_table.insert(),
        values=fill_all_application_data(
            dict(application_identifier="app1"),
            dict(application_identifier="app2"),
            dict(application_identifier="app3"),
        ),
    )
    count = await database.fetch_all("SELECT COUNT(*) FROM applications")
    assert count[0][0] == 3

    inject_security_header("owner1@org.com", Permissions.APPLICATIONS_VIEW)
    response = await client.get("/jobbergate/applications/")
    assert response.status_code == status.HTTP_200_OK

    data = response.json()
    results = data.get("results")
    assert results
    assert sorted([d["application_identifier"] for d in results]) == [
        "app1",
        "app2",
        "app3",
    ]

    pagination = data.get("pagination")
    assert pagination == dict(
        total=3,
        start=None,
        limit=None,
    )


@pytest.mark.asyncio
async def test_get_application___bad_permission(
    client,
    fill_all_application_data,
    inject_security_header,
):
    """
    Test that it is not possible to list applications without proper permission.

    This test proves that the GET /applications returns 403 as status code in the response.
    We show this by making a request with an user without creating the permission, and then asserting the
    status code in the response.
    """
    await database.execute_many(
        query=applications_table.insert(),
        values=fill_all_application_data(
            dict(application_identifier="app1"),
            dict(application_identifier="app2"),
            dict(application_identifier="app3"),
        ),
    )
    count = await database.fetch_all("SELECT COUNT(*) FROM applications")
    assert count[0][0] == 3

    inject_security_header("owner1@org.com", "INVALID_PERMISSION")
    response = await client.get("/jobbergate/applications/")
    assert response.status_code == status.HTTP_403_FORBIDDEN


@pytest.mark.asyncio
async def test_get_applications__with_user_param(
    client,
    fill_all_application_data,
    inject_security_header,
):
    """
    Test applications list doesn't include applications owned by other users with `user` param.

    This test proves that the user making the request cannot see applications owned by other users.
    We show this by creating applications that are owned by another user id and assert that
    the user making the request to list applications doesn't see any of the other user's
    applications in the response
    """
    await database.execute_many(
        query=applications_table.insert(),
        values=fill_all_application_data(
            dict(
                application_identifier="app1",
                application_owner_email="owner1@org.com",
            ),
            dict(
                application_identifier="app2",
                application_owner_email="owner1@org.com",
            ),
            dict(
                application_identifier="app3",
                application_owner_email="owner999@org.com",
            ),
        ),
    )
    count = await database.fetch_all("SELECT COUNT(*) FROM applications")
    assert count[0][0] == 3

    inject_security_header("owner1@org.com", Permissions.APPLICATIONS_VIEW)
    response = await client.get("/jobbergate/applications")
    assert response.status_code == status.HTTP_200_OK

    data = response.json()
    results = data.get("results")
    assert sorted(d["application_identifier"] for d in results) == ["app1", "app2", "app3"]

    pagination = data.get("pagination")
    assert pagination == dict(
        total=3,
        start=None,
        limit=None,
    )

    response = await client.get("/jobbergate/applications?user=true")
    assert response.status_code == status.HTTP_200_OK

    data = response.json()
    results = data.get("results")
    assert sorted(d["application_identifier"] for d in results) == ["app1", "app2"]

    pagination = data.get("pagination")
    assert pagination == dict(
        total=2,
        start=None,
        limit=None,
    )


@pytest.mark.asyncio
async def test_get_applications__with_all_param(
    client,
    fill_all_application_data,
    inject_security_header,
):
    """
    Test that listing applications, when all=True, contains applications without identifiers.

    This test proves that the user making the request can see applications owned by other users.
    We show this by creating three applications, two that are owned by the user making the request, and one
    owned by another user. Assert that the response to GET /applications/?all=True includes all three
    applications.
    """
    await database.execute_many(
        query=applications_table.insert(),
        values=fill_all_application_data(
            dict(application_identifier="app1", application_owner_email="owner1@org.com"),
            dict(application_owner_email="owner1@org.com"),
            dict(application_identifier="app3", application_owner_email="owner999@org.com"),
        ),
    )
    count = await database.fetch_all("SELECT COUNT(*) FROM applications")
    assert count[0][0] == 3

    inject_security_header("owner1@org.com", Permissions.APPLICATIONS_VIEW)

    response = await client.get("/jobbergate/applications")
    assert response.status_code == status.HTTP_200_OK

    data = response.json()
    results = data.get("results")
    assert [d["application_identifier"] for d in results] == ["app1", "app3"]

    pagination = data.get("pagination")
    assert pagination == dict(
        total=2,
        start=None,
        limit=None,
    )

    response = await client.get("/jobbergate/applications/?all=True")
    assert response.status_code == status.HTTP_200_OK

    data = response.json()
    results = data.get("results")
    pagination = data.get("pagination")
    assert pagination == dict(
        total=3,
        start=None,
        limit=None,
    )


@pytest.mark.asyncio
async def test_get_applications__with_search_param(
    client,
    fill_all_application_data,
    inject_security_header,
):
    """
    Test that listing applications, when search=<search terms>, returns matches.

    This test proves that the user making the request will be shown applications that match the search string.
    We show this by creating applications and using various search queries to match against them.

    Assert that the response to GET /applications?search=<search temrms> includes correct matches.
    """
    await database.execute_many(
        query=applications_table.insert(),
        values=fill_all_application_data(
            dict(
                application_name="test name one",
                application_identifier="app1",
                application_owner_email="one@org.com",
            ),
            dict(
                application_name="test name two",
                application_identifier="app2",
                application_owner_email="two@org.com",
            ),
            dict(
                application_name="test name twenty-two",
                application_identifier="app22",
                application_owner_email="twenty-two@org.com",
                application_description="a long description of this application",
            ),
        ),
    )
    count = await database.fetch_all("SELECT COUNT(*) FROM applications")
    assert count[0][0] == 3

    inject_security_header("admin@org.com", Permissions.APPLICATIONS_VIEW)

    response = await client.get("/jobbergate/applications?all=true&search=one")
    assert response.status_code == status.HTTP_200_OK
    data = response.json()
    results = data.get("results")
    assert [d["application_identifier"] for d in results] == ["app1"]

    response = await client.get("/jobbergate/applications?all=true&search=two")
    assert response.status_code == status.HTTP_200_OK
    data = response.json()
    results = data.get("results")
    assert sorted([d["application_identifier"] for d in results]) == ["app2", "app22"]

    response = await client.get("/jobbergate/applications?all=true&search=long")
    assert response.status_code == status.HTTP_200_OK
    data = response.json()
    results = data.get("results")
    assert [d["application_identifier"] for d in results] == ["app22"]

    response = await client.get("/jobbergate/applications?all=true&search=name+test")
    assert response.status_code == status.HTTP_200_OK
    data = response.json()
    results = data.get("results")
    assert sorted([d["application_identifier"] for d in results]) == ["app1", "app2", "app22"]


@pytest.mark.asyncio
async def test_get_applications__with_sort_params(
    client,
    fill_all_application_data,
    inject_security_header,
):
    """
    Test that listing applications with sort params returns correctly ordered matches.

    This test proves that the user making the request will be shown applications sorted in the correct order
    according to the ``sort_field`` and ``sort_ascending`` parameters.
    We show this by creating applications and using various sort parameters to order them.

    Assert that the response to GET /applications?sort_field=<field>&sort_ascending=<bool> includes correctly
    sorted applications.
    """
    await database.execute_many(
        query=applications_table.insert(),
        values=fill_all_application_data(
            dict(application_name="A", application_identifier="Z"),
            dict(application_name="B", application_identifier="Y"),
            dict(application_name="C", application_identifier="X"),
        ),
    )
    count = await database.fetch_all("SELECT COUNT(*) FROM applications")
    assert count[0][0] == 3

    inject_security_header("admin@org.com", Permissions.APPLICATIONS_VIEW)

    response = await client.get("/jobbergate/applications?all=true&sort_field=application_name")
    assert response.status_code == status.HTTP_200_OK
    data = response.json()
    results = data.get("results")
    assert [d["application_identifier"] for d in results] == ["Z", "Y", "X"]

    response = await client.get(
        "/jobbergate/applications?all=true&sort_field=application_name&sort_ascending=false"
    )
    assert response.status_code == status.HTTP_200_OK
    data = response.json()
    results = data.get("results")
    assert [d["application_identifier"] for d in results] == ["X", "Y", "Z"]

    response = await client.get("/jobbergate/applications?all=true&sort_field=application_identifier")
    assert response.status_code == status.HTTP_200_OK
    data = response.json()
    results = data.get("results")
    assert [d["application_identifier"] for d in results] == ["X", "Y", "Z"]

    response = await client.get("/jobbergate/applications?all=true&sort_field=application_config")
    assert response.status_code == status.HTTP_400_BAD_REQUEST
    assert "Invalid sorting column requested" in response.text


@pytest.mark.asyncio
async def test_get_applications__with_pagination(
    client,
    fill_all_application_data,
    inject_security_header,
):
    """
    Test that listing applications works with pagination.

    This test proves that the user making the request can see applications paginated.
    We show this by creating three applications and assert that the response is correctly paginated.
    """
    await database.execute_many(
        query=applications_table.insert(),
        values=fill_all_application_data(
            dict(application_identifier="app1", application_owner_email="owner1@org.com"),
            dict(application_identifier="app2", application_owner_email="owner1@org.com"),
            dict(application_identifier="app3", application_owner_email="owner1@org.com"),
            dict(application_identifier="app4", application_owner_email="owner1@org.com"),
            dict(application_identifier="app5", application_owner_email="owner1@org.com"),
        ),
    )
    count = await database.fetch_all("SELECT COUNT(*) FROM applications")
    assert count[0][0] == 5

    inject_security_header("owner1@org.com", Permissions.APPLICATIONS_VIEW)
    response = await client.get(
        "/jobbergate/applications/?start=0&limit=1&all=true&sort_field=application_identifier"
    )
    assert response.status_code == status.HTTP_200_OK

    data = response.json()
    results = data.get("results")
    assert results
    assert [d["application_identifier"] for d in results] == ["app1"]

    pagination = data.get("pagination")
    assert pagination == dict(
        total=5,
        start=0,
        limit=1,
    )

    response = await client.get("/jobbergate/applications/?start=1&limit=2&all=true")
    assert response.status_code == status.HTTP_200_OK

    data = response.json()
    results = data.get("results")
    assert results
    assert [d["application_identifier"] for d in results] == ["app3", "app4"]

    pagination = data.get("pagination")
    assert pagination == dict(
        total=5,
        start=1,
        limit=2,
    )

    response = await client.get("/jobbergate/applications/?start=2&limit=2&all=true")
    assert response.status_code == status.HTTP_200_OK

    data = response.json()
    results = data.get("results")
    assert results
    assert [d["application_identifier"] for d in results] == ["app5"]

    pagination = data.get("pagination")
    assert pagination == dict(
        total=5,
        start=2,
        limit=2,
    )


@pytest.mark.asyncio
async def test_update_application(
    client,
    fill_application_data,
    inject_security_header,
    time_frame,
):
    """
    Test that an application is updated via PUT.

    This test proves that an application's values are correctly updated following a PUT request to the
    /application/<id> endpoint. We show this by asserting that the values provided to update the
    application are returned in the response made to the PUT /application/<id> endpoint.
    """
    await database.execute(
        query=applications_table.insert(),
        values=fill_application_data(
            application_identifier="old_identifier",
            application_owner_email="owner1@org.com",
            application_description="old description",
        ),
    )
    rows = await database.fetch_all("select id from applications")
    assert len(rows) == 1
    id = rows[0][0]

    inject_security_header("owner1@org.com", Permissions.APPLICATIONS_EDIT)
    with time_frame() as window:
        response = await client.put(
            f"/jobbergate/applications/{id}",
            json=dict(
                application_name="new_name",
                application_identifier="new_identifier",
                application_description="new_description",
            ),
        )
    assert response.status_code == status.HTTP_200_OK

    data = response.json()
    assert data["application_name"] == "new_name"
    assert data["application_identifier"] == "new_identifier"
    assert data["application_description"] == "new_description"

    query = applications_table.select(applications_table.c.id == id)
    result = await database.fetch_one(query)

    assert result is not None
    assert result["application_name"] == "new_name"
    assert result["application_identifier"] == "new_identifier"
    assert result["application_owner_email"] == "owner1@org.com"
    assert result["application_description"] == "new_description"
    assert result["updated_at"] in window


@pytest.mark.asyncio
async def test_update_application_bad_permission(
    client,
    fill_application_data,
    inject_security_header,
):
    """
    Test that it is not possible to update applications without proper permission.

    This test proves that an application's values are not updated following a PUT request to the
    /application/<id> endpoint by a user without permission. We show this by asserting that the status code
    403 is returned and that the application_data is still the same as before.
    """
    inserted_id = await database.execute(
        query=applications_table.insert(),
        values=fill_application_data(
            application_name="old-name",
            application_identifier="old_identifier",
            application_owner_email="owner1@org.com",
            application_description="old description",
        ),
    )
    count = await database.fetch_all("SELECT COUNT(*) FROM applications")
    assert count[0][0] == 1

    inject_security_header("owner1@org.com", "INVALID_PERMISSION")
    response = await client.put(
        f"/jobbergate/applications/{inserted_id}",
        json=dict(
            application_name="new_name",
            application_identifier="new_identifier",
            application_description="new_description",
        ),
    )
    assert response.status_code == status.HTTP_403_FORBIDDEN

    query = applications_table.select(applications_table.c.id == inserted_id)
    result = await database.fetch_one(query)

    assert result is not None
    assert result["application_name"] == "old-name"
    assert result["application_identifier"] == "old_identifier"
    assert result["application_description"] == "old description"


@pytest.mark.asyncio
@mock.patch("jobbergate_api.apps.applications.routers.write_application_files_to_s3")
async def test_upload_file__works_with_small_file(
    mocked_application_writer,
    client,
    inject_security_header,
    fill_application_data,
    tweak_settings,
    make_dummy_file,
    dummy_application_config,
    make_files_param,
):
    """
    Test that a file is uploaded.

    This test proves that an application's file is uploaded by making sure that the
    boto3 put_object method is called once and a 201 status code is returned. It also
    checks to make sure that the application row in the database has
    `application_uploaded` set.
    """
    inserted_id = await database.execute(
        query=applications_table.insert(),
        values=fill_application_data(application_owner_email="owner1@org.com"),
    )
    application = await fetch_instance(inserted_id, applications_table, ApplicationPartialResponse)
    assert not application.application_uploaded

    dummy_file = make_dummy_file("jobbergate.yaml", content=dummy_application_config)
    inject_security_header("owner1@org.com", Permissions.APPLICATIONS_EDIT)
    with tweak_settings(MAX_UPLOAD_FILE_SIZE=600):
        with make_files_param(dummy_file) as files_param:
            response = await client.post(
                f"/jobbergate/applications/{inserted_id}/upload",
                files=files_param,
            )

    assert response.status_code == status.HTTP_201_CREATED
    mocked_application_writer.assert_called_once()

    application = await fetch_instance(inserted_id, applications_table, ApplicationPartialResponse)
    assert application.application_uploaded


@pytest.mark.asyncio
@mock.patch("jobbergate_api.apps.applications.routers.write_application_files_to_s3")
async def test_upload_file__fails_with_413_on_large_file(
    mocked_application_writer,
    client,
    inject_security_header,
    tweak_settings,
    make_dummy_file,
    make_files_param,
):
    """
    Test that upload fails when the files are too large.
    """
    dummy_file = make_dummy_file("dummy.py", size=600 + 200)
    inject_security_header("owner1@org.com", Permissions.APPLICATIONS_EDIT)
    with tweak_settings(MAX_UPLOAD_FILE_SIZE=600):
        with make_files_param(dummy_file) as files_param:
            response = await client.post(
                "/jobbergate/applications/1/upload",
                files=files_param,
            )

    assert response.status_code == status.HTTP_413_REQUEST_ENTITY_TOO_LARGE

    mocked_application_writer.assert_not_called()


@pytest.mark.asyncio
@mock.patch("jobbergate_api.apps.applications.routers.delete_application_files_from_s3")
async def test_delete_file(mocked_application_deleter, client, inject_security_header, fill_application_data):
    """
    Test that a file is deleted.

    This test proves that an application's file is deleted by making sure that the boto3 put_object method
    is called once and a 201 status code is returned.
    """
    inserted_id = await database.execute(
        query=applications_table.insert(),
        values=fill_application_data(application_owner_email="owner1@org.com", application_uploaded=True),
    )
    application: ApplicationPartialResponse = await fetch_instance(
        inserted_id, applications_table, ApplicationPartialResponse
    )
    assert application.application_uploaded

    inject_security_header("owner1@org.com", Permissions.APPLICATIONS_EDIT)
    response = await client.delete(f"/jobbergate/applications/{inserted_id}/upload")

    assert response.status_code == status.HTTP_204_NO_CONTENT
    mocked_application_deleter.assert_called_once_with(inserted_id)

    application = await fetch_instance(inserted_id, applications_table, ApplicationPartialResponse)
    assert not application.application_uploaded



"""
Test the security module.
"""
from datetime import datetime
from unittest.mock import patch

from armasec import TokenPayload

from jobbergate_api import security


def test_get_domain_configs__loads_only_base_settings():
    """Check if the correct domain configuration is loaded when only one domain is provided."""
    with patch.object(security.settings, "ARMASEC_DOMAIN", new="foo.io"):
        with patch.object(security.settings, "ARMASEC_AUDIENCE", new="https://bar.dev"):
            domain_configs = security.get_domain_configs()

    assert len(domain_configs) == 1
    first_config = domain_configs.pop()
    assert first_config.domain == "foo.io"
    assert first_config.audience == "https://bar.dev"


def test_get_domain_configs__loads_admin_settings_if_all_are_present():
    """Check if the correct domain configuration is loaded when two domains are provided."""
    with patch.object(security.settings, "ARMASEC_DOMAIN", new="foo.io"):
        with patch.object(security.settings, "ARMASEC_AUDIENCE", new="https://bar.dev"):
            with patch.object(security.settings, "ARMASEC_ADMIN_DOMAIN", new="admin.io"):
                domain_configs = security.get_domain_configs()

    assert len(domain_configs) == 1
    first_config = domain_configs.pop()
    assert first_config.domain == "foo.io"
    assert first_config.audience == "https://bar.dev"

    with patch.object(security.settings, "ARMASEC_DOMAIN", new="foo.io"):
        with patch.object(security.settings, "ARMASEC_AUDIENCE", new="https://bar.dev"):
            with patch.object(security.settings, "ARMASEC_ADMIN_DOMAIN", new="admin.io"):
                with patch.object(security.settings, "ARMASEC_ADMIN_AUDIENCE", new="https://admin.dev"):
                    with patch.object(security.settings, "ARMASEC_ADMIN_MATCH_KEY", new="foo"):
                        with patch.object(security.settings, "ARMASEC_ADMIN_MATCH_VALUE", new="bar"):
                            domain_configs = security.get_domain_configs()

    assert len(domain_configs) == 2
    (first_config, second_config) = domain_configs
    assert first_config.domain == "foo.io"
    assert first_config.audience == "https://bar.dev"
    assert second_config.domain == "admin.io"
    assert second_config.audience == "https://admin.dev"
    assert second_config.match_keys == dict(foo="bar")


def test_from_token_payload__omits_email_if_not_in_payload():
    """Check if omiting email causes a failure in from_token_payload()."""
    payload = TokenPayload(
        sub="dummy",
        permissions=[],
        exp=datetime.now(),
        azp="idiot",
    )
    identity = security.IdentityClaims.from_token_payload(payload)
    assert identity.email is None



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['jobbergate_api',
 'jobbergate_api.apps',
 'jobbergate_api.apps.applications',
 'jobbergate_api.apps.job_scripts',
 'jobbergate_api.apps.job_submissions',
 'jobbergate_api.tests',
 'jobbergate_api.tests.apps',
 'jobbergate_api.tests.apps.applications',
 'jobbergate_api.tests.apps.job_scripts',
 'jobbergate_api.tests.apps.job_submissions']

package_data = \
{'': ['*']}

install_requires = \
['Jinja2>=3.0.3,<4.0.0',
 'PyYAML>=6.0,<7.0',
 'alembic>=1.7.5,<2.0.0',
 'armasec>=0.11.0,<0.12.0',
 'asyncpg>=0.22.0,<0.23.0',
 'boto3>=1.17.51,<2.0.0',
 'databases[postgresql]>=0.5.5,<0.6.0',
 'email_validator>=1.1.0,<2.0.0',
 'fastapi>=0.68.0,<0.69.0',
 'file-storehouse==0.3.4',
 'loguru>=0.6.0,<0.7.0',
 'passlib>=1.7.2,<2.0.0',
 'python-dotenv>=0.16.0,<0.17.0',
 'python-multipart>=0.0.5,<0.0.6',
 'sendgrid>=6.9.7,<7.0.0',
 'sentry-sdk>=1.5.0,<2.0.0',
 'typer>=0.4.0,<0.5.0',
 'uvicorn>=0.15.0,<0.16.0',
 'yarl>=1.7.2,<2.0.0']

entry_points = \
{'console_scripts': ['dev-tools = dev_tools:app']}

setup_kwargs = {
    'name': 'jobbergate-api',
    'version': '3.2.4a0',
    'description': 'Jobbergate API',
    'long_description': '===============\n Jobbergate API\n===============\n\nJobbergate API',
    'author': 'Omnivector Solutions',
    'author_email': 'info@omnivector.solutions',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/omnivector-solutions/jobbergate',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



"""
Test the storage module.
"""

from jobbergate_api.storage import build_db_url


def test_build_db_url__creates_database_url_from_parts(tweak_settings):
    """
    Provide a test ase for the ``build_db_url()`` function.

    Tests that the build_db_url function computes a database url value from separate
    DATABASE_ settings when the DEPLOY_ENV is not TEST.
    """
    with tweak_settings(
        DEPLOY_ENV="LOCAL",
        DATABASE_USER="built-base-user",
        DATABASE_PSWD="built-base-pswd",
        DATABASE_HOST="built-base-host",
        DATABASE_NAME="built-base-name",
        DATABASE_PORT=8888,
        TEST_DATABASE_USER="built-test-user",
        TEST_DATABASE_PSWD="built-test-pswd",
        TEST_DATABASE_HOST="built-test-host",
        TEST_DATABASE_NAME="built-test-name",
        TEST_DATABASE_PORT=9999,
    ):
        assert build_db_url() == (
            "postgresql://built-base-user:built-base-pswd@built-base-host:8888/built-base-name"
        )


def test_build_db_url__uses_TEST_prefixed_database_settings_if_passed_the_force_test_flag(tweak_settings):
    """
    Provide a test ase for the ``build_db_url()`` function.

    Tests that the build_db_url function computes a database url value from separate
    TEST_DATABASE_ settings when the DEPLOY_ENV is TEST.
    """
    with tweak_settings(
        DEPLOY_ENV="TEST",
        DATABASE_USER="built-base-user",
        DATABASE_PSWD="built-base-pswd",
        DATABASE_HOST="built-base-host",
        DATABASE_NAME="built-base-name",
        DATABASE_PORT=8888,
        TEST_DATABASE_USER="built-test-user",
        TEST_DATABASE_PSWD="built-test-pswd",
        TEST_DATABASE_HOST="built-test-host",
        TEST_DATABASE_NAME="built-test-name",
        TEST_DATABASE_PORT=9999,
    ):
        assert build_db_url() == (
            "postgresql://built-test-user:built-test-pswd@built-test-host:9999/built-test-name"
        )



"""
Instantiates armasec resources for auth on api endpoints using project settings.

Also provides a factory function for TokenSecurity to reduce boilerplate.
"""
from typing import List, Optional

from armasec import Armasec, TokenPayload
from armasec.schemas import DomainConfig
from loguru import logger
from pydantic import BaseModel, EmailStr

from jobbergate_api.config import settings


def get_domain_configs() -> List[DomainConfig]:
    """
    Return a list of DomainConfig objects based on the input variables for the Settings class.
    """
    domain_configs = [
        DomainConfig(
            domain=settings.ARMASEC_DOMAIN,
            audience=settings.ARMASEC_AUDIENCE,
            debug_logger=logger.debug if settings.ARMASEC_DEBUG else None,
        )
    ]
    if all(
        [
            settings.ARMASEC_ADMIN_DOMAIN,
            settings.ARMASEC_ADMIN_AUDIENCE,
            settings.ARMASEC_ADMIN_MATCH_KEY,
            settings.ARMASEC_ADMIN_MATCH_VALUE,
        ]
    ):
        domain_configs.append(
            DomainConfig(
                domain=settings.ARMASEC_ADMIN_DOMAIN,
                audience=settings.ARMASEC_ADMIN_AUDIENCE,
                match_keys={settings.ARMASEC_ADMIN_MATCH_KEY: settings.ARMASEC_ADMIN_MATCH_VALUE},
                debug_logger=logger.debug if settings.ARMASEC_DEBUG else None,
            )
        )
    return domain_configs


guard = Armasec(domain_configs=get_domain_configs())


class IdentityClaims(BaseModel):
    """
    Provide a pydantic data model containing user data extracted from an access token.
    """

    email: Optional[EmailStr]
    client_id: Optional[str]

    @classmethod
    def from_token_payload(cls, payload: TokenPayload) -> "IdentityClaims":
        """
        Create an instance from a Token payload.

        Automatically validates that the email is an email address if it is provided.
        """
        init_kwargs = dict(
            client_id=payload.client_id,
        )
        email = getattr(payload, "email", None)
        if email is not None:
            init_kwargs.update(email=email)
        return cls(**init_kwargs)



"""
Defines the schema for the resource Application.
"""
from datetime import datetime
from textwrap import dedent
from typing import Any, Dict, List, Optional

from pydantic import BaseModel

from jobbergate_api.meta_mapper import MetaField, MetaMapper

application_meta_mapper = MetaMapper(
    id=MetaField(
        description="The unique database identifier for the instance",
        example=101,
    ),
    created_at=MetaField(
        description="The timestamp for when the instance was created",
        example="2021-12-28 23:13:00",
    ),
    updated_at=MetaField(
        description="The timestamp for when the instance was last updated",
        example="2021-12-28 23:52:00",
    ),
    application_name=MetaField(
        description="The unique name of the application",
        example="test-application-88",
    ),
    application_identifier=MetaField(
        description="A human-friendly label used for lookup on frequently accessed applications",
        example="App88",
    ),
    application_owner_email=MetaField(
        description="A long-form textual description of the application",
        example="tucker@omnivector.solutions",
    ),
    application_file=MetaField(
        description="The source code for the application file content text",
        example=dedent(
            """
            from jobbergate_cli.application_base import JobbergateApplicationBase
            from jobbergate_cli import appform


            class JobbergateApplication(JobbergateApplicationBase):

                def mainflow(self, data):
                    questions = []

                    questions.append(appform.List(
                        variablename="partition",
                        message="Choose slurm partition:",
                        choices=self.application_config['partitions'],
                    ))

                    questions.append(appform.Text(
                        variablename="job_name",
                        message="Please enter a jobname",
                        default=self.application_config['job_name']
                    ))
                    return questions
            """
        ).strip(),
    ),
    application_config=MetaField(
        description="Application config file content (.yaml) as text",
        example=dedent(
            """
            application_config:
              job_name: rats
              partitions:
              - juju-compute-SCqp
            jobbergate_config:
              default_template: test_job_script.sh
              output_directory: .
              supporting_files:
              - test_job_script.sh
              supporting_files_output_name:
                test_job_script.sh:
                - support_file_b.py
              template_files:
              - templates/test_job_script.sh
            """
        ).strip(),
    ),
    application_uploaded=MetaField(
        description="Indicates if the application file zip has been uploaded yet.",
        example=True,
    ),
)


class JobbergateConfig(BaseModel):
    """
    Model for Jobbergate configuration (subsection at the yaml file).
    """

    default_template: str
    supporting_files: Optional[List[str]]
    supporting_files_output_name: Optional[Dict[str, List[str]]]
    template_files: Optional[List[str]]
    job_script_name: Optional[str]
    output_directory: str = "."

    class Config:
        extra = "allow"


class ApplicationConfig(BaseModel):
    """
    Model for application configuration, used to parse the yaml file.
    """

    application_config: Dict[str, Any]
    jobbergate_config: JobbergateConfig


class ApplicationCreateRequest(BaseModel):
    """
    Request model for creating Application instances.
    """

    application_name: str
    application_identifier: Optional[str]
    application_description: Optional[str] = None

    class Config:
        schema_extra = application_meta_mapper


class ApplicationUpdateRequest(BaseModel):
    """
    Request model for updating Application instances.
    """

    application_name: Optional[str]
    application_identifier: Optional[str]
    application_description: Optional[str]
    application_config: Optional[str]

    class Config:
        schema_extra = application_meta_mapper


class ApplicationPartialResponse(BaseModel):
    """
    Complete model to match database for the Application resource.
    """

    id: int
    created_at: Optional[datetime] = datetime.utcnow()
    updated_at: Optional[datetime] = datetime.utcnow()
    application_name: str
    application_identifier: Optional[str]
    application_description: Optional[str]
    application_owner_email: str
    application_uploaded: bool

    class Config:
        orm_mode = True
        schema_extra = application_meta_mapper


class ApplicationResponse(ApplicationPartialResponse):
    """
    Complete model to match database for the Application resource.
    """

    application_config: Optional[str]
    application_source_file: Optional[str]
    application_templates: Optional[Dict[str, str]]

    class Config:
        orm_mode = True
        schema_extra = application_meta_mapper



"""
JobSubmission resource schema.
"""
from datetime import datetime
from pathlib import Path
from typing import Optional

from pydantic import BaseModel, Extra, Field

from jobbergate_api.apps.job_submissions.constants import JobSubmissionStatus
from jobbergate_api.meta_mapper import MetaField, MetaMapper

job_submission_meta_mapper = MetaMapper(
    id=MetaField(
        description="The unique database identifier for the instance",
        example=101,
    ),
    created_at=MetaField(
        description="The timestamp for when the instance was created",
        example="2021-12-28 23:13:00",
    ),
    updated_at=MetaField(
        description="The timestamp for when the instance was last updated",
        example="2021-12-28 23:52:00",
    ),
    job_submission_name=MetaField(
        description="The unique name of the job submission",
        example="test-job-submission-77",
    ),
    job_submission_description=MetaField(
        description="A text field providing a human-friendly description of the job_submission",
        example="Submission for the Foo job on sample 99 using the bar variant",
    ),
    job_submission_owner_email=MetaField(
        description="The email of the owner/creator of the instance",
        example="tucker@omnivector.solutions",
    ),
    job_script_id=MetaField(
        description="The foreign-key to the job_script from which this instance was created",
        example=71,
    ),
    execution_duirectory=MetaField(
        description="The directory on the cluster where the job should be executed",
        example="/home/someuser/job-data/test-77",
    ),
    slurm_job_id=MetaField(
        description="The id for the slurm job executing this job_submission",
        example="1883",
    ),
    client_id=MetaField(
        description="The client_id of the cluster where this job submission should execute",
        example="mega-cluster-1",
    ),
    status=MetaField(
        description=f"The status of the job submission. Must be one of {JobSubmissionStatus.pretty_list()}",
        example=JobSubmissionStatus.CREATED,
    ),
    report_message=MetaField(
        description="The report message received from cluster-agent when a job submission is rejected",
        example="Unrecognized SBATCH arguments",
    ),
)


class JobSubmissionCreateRequest(BaseModel):
    """
    Request model for creating JobSubmission instances.
    """

    job_submission_name: str
    job_submission_description: Optional[str]
    job_script_id: int
    execution_directory: Optional[Path]
    client_id: Optional[str]

    class Config:
        schema_extra = job_submission_meta_mapper


class JobSubmissionUpdateRequest(BaseModel):
    """
    Request model for updating JobSubmission instances.
    """

    job_submission_name: Optional[str]
    job_submission_description: Optional[str]
    execution_directory: Optional[Path]
    status: Optional[JobSubmissionStatus]

    class Config:
        schema_extra = job_submission_meta_mapper


class JobSubmissionResponse(BaseModel):
    """
    Complete model to match the database for the JobSubmission resource.
    """

    id: Optional[int] = Field(None)
    created_at: Optional[datetime]
    updated_at: Optional[datetime]
    job_submission_name: str
    job_submission_description: Optional[str]
    job_submission_owner_email: str
    job_script_id: int
    execution_directory: Optional[Path]
    slurm_job_id: Optional[int]
    client_id: Optional[str]
    status: JobSubmissionStatus
    report_message: Optional[str]

    class Config:
        orm_mode = True
        schema_extra = job_submission_meta_mapper


class PendingJobSubmission(BaseModel, extra=Extra.ignore):
    """
    Specialized model for the cluster-agent to pull pending job_submissions.

    Model also includes data from its job_script and application sources.
    """

    id: Optional[int] = Field(None)
    job_submission_name: str
    job_submission_owner_email: str
    execution_directory: Optional[Path]
    job_script_name: str
    job_script_data_as_string: str
    application_name: str


class ActiveJobSubmission(BaseModel, extra=Extra.ignore):
    """
    Specialized model for the cluster-agent to pull an active job_submission.
    """

    id: int
    job_submission_name: str
    slurm_job_id: int



"""
Database model for the Application resource.
"""
from sqlalchemy import Boolean, DateTime, Integer, String, Table
from sqlalchemy.sql import False_, func
from sqlalchemy.sql.schema import Column

from jobbergate_api.metadata import metadata

applications_table = Table(
    "applications",
    metadata,
    Column("id", Integer, primary_key=True),
    Column("application_name", String, nullable=False, index=True),
    Column("application_identifier", String, unique=True, index=True),
    Column("application_description", String, default=""),
    Column("application_owner_email", String, nullable=False, index=True),
    Column("application_config", String, nullable=True),
    Column("application_uploaded", Boolean, nullable=False, default=False_()),
    Column("created_at", DateTime, nullable=False, default=func.now()),
    Column("updated_at", DateTime, nullable=False, default=func.now(), onupdate=func.now()),
)

searchable_fields = [
    applications_table.c.application_name,
    applications_table.c.application_identifier,
    applications_table.c.application_description,
    applications_table.c.application_owner_email,
]

sortable_fields = [
    applications_table.c.id,
    applications_table.c.application_name,
    applications_table.c.application_identifier,
    applications_table.c.application_owner_email,
    applications_table.c.application_uploaded,
    applications_table.c.created_at,
    applications_table.c.updated_at,
]
